@@ -1,7 +1,9 @@
+import copy
 import networkx
 from networkx import MultiDiGraph
 from typing import List, Dict
 from Bio import Phylo
+from Bio.Phylo import Newick
 from io import StringIO
 from pthr_db_caller.models.panther import NodeDatFile
 
@@ -20,7 +22,7 @@ class PantherTreePhylo:
             tree_phylo = next(Phylo.parse(tree_string, "newick"))
             # Leaves parse clean due to not having species name in 'S:'
 
-        self.tree = tree_phylo
+        self.tree: Newick.Tree = tree_phylo
 
 
 def extract_clade_name(clade_comment):
@@ -55,6 +57,7 @@ class PantherTreeGraph:
         self.graph = MultiDiGraph()
         self.name: str = tree_name
         self.ptn_to_an: Dict = None
+        self.phylo: PantherTreePhylo = None
 
     # Recursive method to fill graph from Phylo clade, parsing out node accession and species name (if present)
     def add_children(self, parent_clade):
@@ -113,17 +116,75 @@ class PantherTreeGraph:
 
         # Parse Newick line
         phylo = PantherTreePhylo(tree_file)
-        # Fill networkx graph from Phylo obj
-        pthr_tree_graph.add_children(phylo.tree.clade)
+        pthr_tree_graph.init_from_phylo(phylo)
         # Fill in long IDs on leaf nodes
         pthr_tree_graph.extract_leaf_ids(tree_file)
         # Fill in PTNs if node_file specified
         if node_file:
             pthr_tree_graph.extract_node_properties(node_file)
 
-
         return pthr_tree_graph
 
+    def init_from_phylo(self, phylo: PantherTreePhylo):
+        self.phylo: PantherTreePhylo = phylo
+        # Fill networkx graph from Phylo obj
+        self.add_children(phylo.tree.clade)
+
+    def write(self, outpath):
+        transformed_tree = copy.deepcopy(self.phylo.tree)
+        self.traverse(transformed_tree.clade)
+        Phylo.write(transformed_tree, outpath, 'newick')
+
+    def traverse(self, c, parent_species=None):
+        species, nid = extract_clade_name(c.comment)
+        if nid == "":
+            # Likely a leaf node
+            cn = self.node(c.name)
+            long_id = cn.get("long_id")
+            species = long_id.split("|")[0]
+            nid = long_id.split("|")[2].split("=")[1]
+            if species == "":
+                species = parent_species
+            c.name = self.newick_name_fmt(species, nid)
+        else:
+            c.name = "1"
+        # Transform event type values ("0>1" -> "S", "1>0" -> "D")
+        if c.comment:
+            # &&NHX:Ev=0>1:S=Amoebozoa:ID=AN7
+            new_comment_elements = ["&&NHX"]
+            if "Ev=0>1" in c.comment:
+                new_comment_elements.append("Ev=S")  # speciation
+            elif "Ev=1>0" in c.comment:
+                new_comment_elements.append("Ev=D")  # duplication
+            elif "Ev=0>0" in c.comment:
+                new_comment_elements.append("Ev=D")  # horizontal transfer, but pretend like it's duplication
+            c.comment = ":".join(new_comment_elements)
+        for child_clade in c.clades:
+            self.traverse(child_clade, parent_species=species)
+
+    def prune_species(self, taxon_list: List):
+        """
+        Process:
+        1. Find all terminal leaf nodes having taxon (OS code format) NOT in taxon_list. Delete these.
+        2. Starting from the terminal nodes, traverse up, deleting all parent nodes until reaching a node with >1 children.
+        3. Stop traversing that path and move on to traversing up the next terminal node
+        :param taxon_list: The List of OS codes ('5-letter' OSCODES. Ex: HUMAN, MOUSE, SCHPO) that will remain in tree
+        after pruning list non-members
+        """
+        for leaf in self.leaves():
+            long_id = self.node(leaf)["long_id"]
+            species = long_id.split("|")[0]
+            if species not in taxon_list:
+                self.prune_up(leaf)
+
+    @staticmethod
+    def newick_name_fmt(species, nid):
+        # Name formats
+        # LUCA_AN0 (internal speciation)
+        # SALTY_AN159 (internal duplication)
+        # ENTHI_C4M9L1 (leaf)
+        return "_".join([species, nid])
+
     def node(self, node):
         return self.graph.node.get(node)
 
@@ -176,5 +237,24 @@ class PantherTreeGraph:
         ancestors_of_desc = self.ancestors(descendant_node)
         return list(set(descendants_of_anc) & set(ancestors_of_desc))
 
+    def prune_up(self, node):
+        """
+        remove_node(node) and, if node.parent has multiple children, prune_up(node.parent)
+        :param node:
+        :return:
+        """
+        parents = self.parents(node)
+        if len(parents) > 1:
+            print("WARNING: Tree node {} has multiple parents:".format(node), parents)
+        if len(self.children(node)) < 2:
+            self.prune_up(parents[0])
+            self.remove_node(node)
+
+    def remove_node(self, node):
+        # Here, we remove from both graph and phylo
+        self.graph.remove_node(node)
+        if self.phylo.tree.find_any(node):
+            self.phylo.tree.prune(node)
+
     def __len__(self):
         return len(self.graph)
